/* ltmodem.c: pci interface & environment for core
 * 	
 * author: Aleksey Chentsov
 * copying: LGPL
 */

#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/pci.h>

#include "v8250.h"
#include "core.h"

#define MARTIAN_DEV_TOSHIBA	(1 << 0)
#define MPREFIX			__stringify (KBUILD_MODNAME)

#define MERROR(fmt...)		printk (KERN_ERR MPREFIX ": " fmt)

#define MARTIAN_DEBUG

#ifdef MARTIAN_DEBUG
#  define ASSERT(cond, action)	do {if (! (cond)) { action }} while (0)
#else
#  define ASSERT(cond, action) 	do {} while (0)
#endif 	


struct martian_metrics {
	unsigned BaseAddress;
	unsigned CommAddress;

	unsigned int irq;

	unsigned short vendor, device;
	unsigned short subsystem_vendor;
	unsigned short subsystem_device;
	unsigned flags;
};

struct _ltmodem {
	spinlock_t lock;
	unsigned int irq;
	unsigned int line;
	struct vuart_host host;
	struct delayed_work work;
} ltmodem;

unsigned int country_id = 0x19;

/* vuart interface */
static unsigned int serial_in(struct vuart_host *host, int offset)
{
	return read_vuart_port(offset);
}

static void serial_out(struct vuart_host *host, int offset, int value)
{
	write_vuart_port(offset, value);
}

void dump_core_vars(void );
static void serial_startup(struct vuart_host *host)
{
	vxdPortOpen();
	dump_core_vars();
}

static void serial_shutdown(struct vuart_host *host)
{
	vxdPortClose();
}

struct vuart_host_ops lt_uart_ops = {
	.in	  = serial_in,
	.out	  = serial_out,
	.startup  = serial_startup,
	.shutdown = serial_shutdown,
	.sleep	  = NULL
};

static unsigned int init_vport(struct _ltmodem *ltdev)
{
	struct vuart_config conf = {
		.type	= VUART_GENERIC,
		.flags	= VPF_INTERRUPT
	};
	
	ltdev->host.ops = &lt_uart_ops; 

	return serial_v8250_register_port(&conf, &ltdev->host);
}


/* hardware */
static irqreturn_t ltmodem_isr(int irq, void *_dev)
{
	struct _ltmodem *dev = (struct _ltmodem *)_dev;
	int handled = 1;

	spin_lock (&dev->lock);
	dp_dsp_isr();
	spin_unlock (&dev->lock);

	return IRQ_RETVAL(handled);
}

asmlinkage u8 dp_regread(u8 reg)
{
	unsigned long flags;
	u8 val;

	spin_lock_irqsave(&ltmodem.lock, flags);
	val = dp_regread_nonint(reg);
	spin_unlock_irqrestore(&ltmodem.lock, flags);

	return val;
}

asmlinkage void dp_regwrite(u8 reg, u8 val)
{
	unsigned long flags;

	spin_lock_irqsave(&ltmodem.lock, flags);
	dp_regwrite_nonint(reg, val);
	spin_unlock_irqrestore(&ltmodem.lock, flags);
}

/* redone ticker */
/* 
 * timer_tick takes from usec to hundreds msec 
 * which is not affordable in timer bh
 * since we have no straightforward way to distinguish how much would it be
 * worst is assumed and implementation with workitem 
 */

//void work_func(void *arg)
void work_func(struct work_struct *arg)
{
	/* assert(VMODEM_Timer_Active) */
	if (VMODEM_Timer_Active) {
		UART_background();

		if (VMODEM_Timer_Active)
			UART_background();

		if (VMODEM_Timer_Active) {
			/* conditional ? */
			struct _ltmodem *ltdev = (struct _ltmodem *)arg;
			serial_v8250_interrupt(ltmodem.line);
		}

		if (VMODEM_Timer_Active)
			schedule_delayed_work (&ltmodem.work, HZ/100);
	}
}

void lt_init_timer(void)
{}

asmlinkage
void lt_add_timer(void (asmlinkage *timerfunction)(unsigned long))
{
	schedule_delayed_work (&ltmodem.work, HZ/100);
}

#define	DUMP_CORE_VAR(var)	pr_info("martian: \t" #var "\t= %d\n", var)
void dump_core_vars(void )
{
	pr_info("martian: core vars:\n");
	pr_info("martian: \ts_x82\t= %x\n", *(u16 *)(&S[0x82]));
	pr_info("martian: \ts_x84\t= %x\n", *(u16 *)(&S[0x84]));
	pr_info("martian: \ts_x86\t= %x\n", *(u16 *)(&S[0x86]));
	pr_info("\ts_x88\t= %x\n", *(u16 *)(&S[0x88]));
	DUMP_CORE_VAR(x_dsp_mars);
	DUMP_CORE_VAR(x_dsp_mars3);
	DUMP_CORE_VAR(BaseAddress);
	DUMP_CORE_VAR(ComAddress);
	DUMP_CORE_VAR(Irq);
	DUMP_CORE_VAR(S[0x7e]);
}

static void init_core(struct martian_metrics *mconf)
{
	InitVariables();

	BaseAddress 	= mconf->BaseAddress;
	ComAddress 	= mconf->CommAddress;
	Irq 		= mconf->irq;

	ll_set_pci_id (
		(mconf->flags & MARTIAN_DEV_TOSHIBA) ? 0x440 : mconf->device, 
		(mconf->flags & MARTIAN_DEV_TOSHIBA) ? 0x11c1 : mconf->vendor, 
		mconf->subsystem_device,
		mconf->subsystem_vendor
		);

	{
		extern char eeprom[];
		eeprom[0] = country_id;
	}

	lucent_init_modem();
	dump_core_vars();
}

struct _ltmodem *ltmodem_add (struct martian_metrics *metrics, int *error)
{
	struct _ltmodem *dev = &ltmodem;
	int ret;

	dev->irq = metrics->irq;

	// as soon interrupt allocated we should be ready to handle it
	spin_lock_init(&dev->lock);

	// claim resource
	ret = request_irq(
		dev->irq, 
		ltmodem_isr,  
		IRQF_DISABLED | IRQF_SHARED, 
		"ltmodem", 
		dev
	);

	if (ret < 0) {
		MERROR ("failed to request irq\n");
		goto err;
	}

	INIT_DELAYED_WORK(&dev->work, work_func);

	init_core(metrics); //, dev);
	dev->line = init_vport(dev);
	
	pr_info("martian: added device %x:%x " 
			"BaseAddress = 0x%x, CommAddres = 0x%x, irq = %d\n", 
			metrics->vendor, metrics->device, 
			metrics->BaseAddress, metrics->CommAddress,
			metrics->irq
		);

	return dev;

err:
	*error = ret;
	return NULL;
}

static int __devinit 
martian_device_gatherinfo (const struct pci_dev *dev, const struct pci_device_id *ent, struct martian_metrics *metrics)
{
	int bar;

	memset(metrics, 0, sizeof *metrics);

	/*** common setting ***/
	metrics->subsystem_device = dev->subsystem_device;
	metrics->subsystem_vendor = dev->subsystem_vendor;

	metrics->irq = dev->irq;

	if (ent && ent->subvendor == PCI_VENDOR_ID_ATT) {
		/* Toshibas */

		switch (dev->class >> 8) {
		case PCI_CLASS_COMMUNICATION_SERIAL:
		case PCI_CLASS_COMMUNICATION_MULTISERIAL:
		case PCI_CLASS_COMMUNICATION_MODEM:
		case PCI_CLASS_COMMUNICATION_OTHER:
			break;

		default:
			return -ENODEV;
		}

		if (PCI_FUNC (dev->devfn) != 0) {
			MERROR ("Toshiba outside the range\n");
			return -ENODEV;
		}

		metrics->device = 0x440;
		metrics->vendor = PCI_VENDOR_ID_ATT;

		metrics->CommAddress = pci_resource_start (dev, 1);
		metrics->BaseAddress = pci_resource_start (dev, 2);
		metrics->flags |= MARTIAN_DEV_TOSHIBA;
		pr_debug("martian: Toshiba device\n");

		return 0;
	}

	// check range for xircom
	if (ent && ent->vendor == PCI_VENDOR_ID_XIRCOM 
				&& ent->device == PCI_ANY_ID)

		switch (dev->class >> 8) {
		case PCI_CLASS_COMMUNICATION_SERIAL:
		case PCI_CLASS_COMMUNICATION_MULTISERIAL:
		case PCI_CLASS_COMMUNICATION_MODEM:
		case PCI_CLASS_COMMUNICATION_OTHER:
			if (! (dev->device <= 0x3ff)) {
				MERROR ("XIRCOM outside the range\n");
				return -ENODEV;
			}
			break;

		default:
			return -ENODEV;
		}
	
	// all ids added through sysfs come here

	/* Base/Common Addresses */
	for (bar = 5; bar >= 0; bar--) 
		if ( pci_resource_start (dev, bar) != 0 && 
			(pci_resource_flags (dev, bar) & IORESOURCE_IO) ) {
			metrics->BaseAddress =  pci_resource_start (dev, bar);
			break;
		}

	// continue
	for (bar--; bar >= 0; bar--) 
		if ( pci_resource_start (dev, bar) != 0 && 
			(pci_resource_flags (dev, bar) & IORESOURCE_IO) ) {
			metrics->CommAddress =  pci_resource_start (dev, bar);
			break;
		}

	if (! metrics->BaseAddress) {
		MERROR ("base address not found\n");
		return -ENODEV;
	}

	metrics->device = dev->device;
	metrics->vendor = dev->vendor;

	return 0;
}

static int __devinit 
mars_device_probe (struct pci_dev *dev, const struct pci_device_id *ent)
{
	static int selected = 0;
	int ret;
	struct _ltmodem *ltdev;
	struct martian_metrics metrics;

	if (selected) 
		return -ENODEV;

	ret = pci_enable_device (dev);
	if (ret != 0) {
		MERROR ("failed to enable device\n");
		pr_debug("martian: res %d\n", ret);
		return -ENODEV;
	}

	ret = martian_device_gatherinfo (dev, ent, &metrics);
	if (ret < 0) {
		MERROR ("failed to gather device info\n");
		goto err_disable_pci;
	}

	ltdev = ltmodem_add (&metrics, &ret);
	if (! ltdev) {
		goto err_disable_pci;
	}

	pci_set_drvdata (dev, ltdev);

	selected = 1;

	return 0;

err_disable_pci:
	pci_disable_device (dev);

	return ret;
}

static void __devexit mars_remove (struct pci_dev *dev) 
{
	struct _ltmodem *ltdev = (pci_get_drvdata(dev));
	if (ltdev) {
		serial_v8250_unregister_port(ltdev->line);
		free_irq(ltdev->irq, ltdev);
	}
	pci_disable_device(dev);
}

#include "martian_ids.c"

// the device
struct pci_driver ltmodem_driver = {
	.name		= "ltmodem",
	.probe		= mars_device_probe,
	.remove		= __devexit_p (mars_remove),
	.id_table	= ltmodem_ids
};

/***** kernel module interface *****/

extern void detach_from_serial(struct pci_device_id *ids); 

static int __init ltmodem_init (void) 
{
	pr_info("ltmodem loaded\n");
	detach_from_serial(ltmodem_ids);
	return pci_register_driver (&ltmodem_driver);
}

static void __exit ltmodem_exit(void) 
{
	pci_unregister_driver (&ltmodem_driver);
}

module_exit (ltmodem_exit);
module_init (ltmodem_init);

/***** module info *****/

MODULE_LICENSE ("mixed");
MODULE_DESCRIPTION ("LSI pci modem driver");
MODULE_DEVICE_TABLE( pci, ltmodem_ids );
